-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_import_csv"

-- **************************************************
-- General information about this script
-- **************************************************

FA_import_csv = {}

function FA_import_csv:Name()
    return "Import CSV"
end

function FA_import_csv:Version()
    return "1.4"
end

function FA_import_csv:Description()
    return "Import TVPaint's CSV layered animation export format"
end

function FA_import_csv:Creator()
    return "Fazek"
end

function FA_import_csv:UILabel()
    return("Import CSV animation...")
end

-- **************************************************
-- The guts of this script
-- **************************************************

FA_import_csv.filePath = ""
FA_import_csv.line = ""
FA_import_csv.lpos = 1
FA_import_csv.sep = string.byte(",")
FA_import_csv.dirSep = "\\"

function FA_import_csv:OnMouseDown(moho, mouseEvent)
    self:Run(moho)
end

function FA_import_csv:nextItem()
    local i = self.lpos
    local c
    local prn= string.byte("\"")
    local spc= string.byte(" ")
    local tab= string.byte("\t")
    local result = ""
    local len = self.line:len()

    --trailing spaces
    while i <= len do
        c = self.line:byte(i)
        if (c ~= spc and c ~= tab) then
            break
        end
        i = i + 1
    end
    if i > len then
        self.lpos = 0
        return nil
    end

    if c == prn then
        -- with parentheses

        i = i + 1
        local j = i
        while j <= len do
            c = self.line:byte(j)
            if c == prn then
                if j > i then
                    result = result..self.line:sub(i,j - 1)
                end

                if (self.line:byte(j + 1) == prn) then
                    --double quote: escape sequence
                    --replace with '
                    result = result.."'"
                    i = j + 2
                    j = i
                else
                    break
                end
            else
                j = j + 1
            end
        end

        while j <= len do
            c = self.line:byte(j)
            j = j + 1
            if c == self.sep then
                break
            end
        end
        self.lpos = j
    else
        -- no parentheses

        local j = i

        while j <= len do
            c = self.line:byte(j)
            if c == self.sep then
                break
            end
            j = j + 1
        end
        self.lpos = j + 1
        j = j - 1
        --leading spaces
        while j >= i do
            c = result:byte(j)
            if c ~= spc and c ~= tab then
                break
            end
            j = j - 1
        end
        if j >= i then
            result = self.line:sub(i,j)
        end
    end

    return result
end

function FA_import_csv.hasChild(switchLayer, item)
    local len = switchLayer:CountLayers()
    local i = 0

    while i < len do
        if switchLayer:Layer(i):Name() == item then
            return true
        end
        i = i + 1
    end
    return false
end

function FA_import_csv:Run(moho)
    local setDoc = LM.GUI.Alert(LM.GUI.ALERT_QUESTION,
      "Importing TVPaint's CSV file format", "Shall I import document parameters",
      "(field size, fps and duration) together with the animation?", "Import", "Don't import", "Cancel")

    if setDoc < 0 or setDoc > 1 then
        return
    end

    FA_import_csv.filePath = LM.GUI.OpenFile("Select CSV file to load")
    if self.filePath == "" then
        return
    end

    if self.filePath:sub(-4):lower() ~= ".csv" then
        LM.GUI.Alert(LM.GUI.ALERT_WARNING,"Please select a filename with \".csv\" extension")
        return
    end

    local f = io.open(self.filePath, "r")

    if f == nil then
        LM.GUI.Alert(LM.GUI.ALERT_WARNING,"Sorry, file not found.", self.filePath)
        return
    end

    local pos = self.filePath:len()
    local c

    while pos > 0 do
        c = self.filePath:sub(pos,pos)

        if c == "/" or c == "\\" then
            self.dirSep = c
            break
        end
        pos = pos - 1
    end
    local oldPath = self.filePath:sub(1,-4).."frames"..c
    local basePath = self.filePath:sub(1,pos)

    local step = 0
    local namePos = -1
    local name = "Untitled"
    local widthPos = -1
    local width = moho.document:Width()
    local heightPos = -1
    local height = moho.document:Height()
    local fpsPos = -1
    local fps = moho.document:Fps()
    local layers = {}
    local layerNum = 0
    local cr = string.byte("\r")
    local lf = string.byte("\n")
    local hsh = string.byte("#")
    local rootGroup = nil
    local hasFolders = false

    self.sep = string.byte(",")

    moho.document:PrepUndo(nil)
    moho.document:SetDirty()

    local initFrame = moho.frame
    local frame = 1

    for line in f:lines() do
        pos = line:len()

        while pos >= 0 do
            c = line:byte(pos)
            if c ~= cr and c ~= lf then
                break
            end
            pos = pos - 1
        end

        if pos >= 0 then
            self.line = line:sub(1,pos)
            self.lpos = 1

            if step == 0 then
                if line:len() > 4 and line:sub(1,5) == "UTF-8" then
                    self.sep = line:byte(6)
                end
                step = 1

            elseif step == 1 then
                pos = 0
                while self.lpos > 0 do
                    local item = self:nextItem()

                    if item == nil then
                        break
                    end

                    if item == "Project Name" then
                        namePos = pos
                    elseif item == "Width" then
                        widthPos = pos
                    elseif item == "Height" then
                        heightPos = pos
                    elseif item == "Frame Rate" then
                        fpsPos = pos
                    end
                    pos = pos + 1
                end
                step = 2

            elseif step == 2 then
                pos = 0
                while self.lpos > 0 do
                    local item = self:nextItem()

                    if item == nil then
                        break
                    end

                    if pos == namePos then
                        name = item
                    elseif pos == widthPos then
                        width = item
                    elseif pos == heightPos then
                        height = item
                    elseif pos == fpsPos then
                        fps = item
                    end
                    pos = pos + 1
                end
                step = 3

            else
                local item = self:nextItem()

                if item ~= nil and item:byte(1) == hsh then
                    pos = 0

                    if step ~= 5 then
                        if step == 3 then
                            if setDoc == 0 then
                                moho.document:SetShape(width,height)
                                moho.document:SetFps(fps)
                            end
                            rootGroup = moho:LayerAsGroup( moho:CreateNewLayer(MOHO.LT_GROUP, false) )
                            rootGroup:SetName(name)

                            while self.lpos > 0 do
                                item = self:nextItem()

                                if item == nil then
                                    break
                                end
                                local switchLayer = moho:LayerAsSwitch( moho:CreateNewLayer(MOHO.LT_SWITCH, false) )
                                switchLayer:SetInterpMode(false)

                                layers[pos] = {}
                                layers[pos].switch = switchLayer
                                layers[pos].fileName = nil
                                layers[pos].fullPath = ""
                                moho:PlaceLayerInGroup(switchLayer, rootGroup, false)
                                pos = pos + 1
                            end
                            layerNum = pos
                            self.lpos = 1
                            step = 4
                            item = self:nextItem()
                            pos = 0
                        end

                        if item == "#Layers" then
                            while self.lpos > 0 and pos < layerNum do
                                item = self:nextItem()

                                if item == nil then
                                    break
                                end
                                layers[pos].switch:SetName(item)
                                pos = pos + 1
                            end
                        elseif item == "#Density" then
                            while self.lpos > 0 and pos < layerNum do
                                item = self:nextItem()

                                if item == nil then
                                    break
                                end
                                layers[pos].switch.fAlpha:SetValue(0,item)
                                pos = pos + 1
                            end
                        elseif item == "#Blending" then
                            while self.lpos > 0 and pos < layerNum do
                                item = self:nextItem()

                                if item == nil then
                                    break
                                end
                                local blend = MOHO.BM_NORMAL

                                if item == "Multiply" then
                                    blend = MOHO.BM_MULTIPLY
                                elseif item == "Screen" then
                                    blend = MOHO.BM_SCREEN
                                elseif item == "Overlay" then
                                    blend = MOHO.BM_OVERLAY
                                elseif item == "Add" then
                                    blend = MOHO.BM_ADD
                                elseif item == "Difference" then
                                    blend = MOHO.BM_DIFFERENCE
                                elseif item == "Hue" then
                                    blend = MOHO.BM_HUE
                                elseif item == "Saturation" then
                                    blend = MOHO.BM_SATURATION
                                elseif item == "Value" then
                                    blend = MOHO.BM_LUMINOSITY
                                end

                                layers[pos].switch:SetBlendingMode(blend)
                                pos = pos + 1
                            end
                        elseif item == "#Visible" then
                            while self.lpos > 0 and pos < layerNum do
                                item = self:nextItem()

                                if item == nil then
                                    break
                                end

                                local visible = false

                                if tonumber(item) ~= 0 then
                                    visible = true
                                end
                                layers[pos].switch:SetVisible(visible)
                                pos = pos + 1
                            end
                        elseif item == "#Folder" then
                            hasFolders= true
                            while self.lpos > 0 and pos < layerNum do
                                item = self:nextItem()

                                if item == nil then
                                   break
                                end
                                c = item:sub(-1)
                                if c == "/" or c == "\\" then
                                    item = item:sub(1,-2)
                                end
                                layers[pos].fullPath = item..self.dirSep
                                pos = pos + 1
                            end
                        elseif tonumber(item:sub(2,2)) then
                            -- numeric value, frames are coming
                            self.lpos = 1
                            step = 5
                            item = self:nextItem()
                        end
                    end

                    if step == 5 then
                        while self.lpos > 0 and pos < layerNum do
                            item = self:nextItem()

                            if (item == nil) then
                                break
                            end

                            if item ~= layers[pos].fileName then
                                layers[pos].fileName = item
                                local switchLayer = layers[pos].switch

                                if item == "" then
                                    --blank
 
                                    if not self.hasChild(switchLayer,"-") then
                                        local blankLayer = moho:LayerAsVector(moho:CreateNewLayer(MOHO.LT_VECTOR, false))
                                        blankLayer:SetName("-")
                                        moho:PlaceLayerInGroup(blankLayer,switchLayer,true)
                                    end
                                    layers[pos].switch:SetValue(frame, "-")
                                else
                                    --image frame

                                    if not self.hasChild(switchLayer,item) then
                                        local imageLayer = moho:LayerAsImage(moho:CreateNewLayer(MOHO.LT_IMAGE, false))
                                        moho:PlaceLayerInGroup(imageLayer, switchLayer, false)
                                        local path = nil

                                        if hasFolders then
                                            local tryPath = layers[pos].fullPath
                                            path = tryPath..item
                                            local tryFile = io.open(path,"r")
                                            if tryFile then
                                                tryFile:close()
                                            else
                                                local n = tryPath:len() - 1
                                                local prevPath = ""

                                                while n > 0 do
                                                    local m = n
                                                    while n > 0 do
                                                        c = tryPath:sub(n,n)
                                                        if c == "/" or c == "\\" then
                                                            break
                                                        end
                                                        n = n - 1
                                                    end
                                                    prevPath = tryPath:sub(n + 1,m)..self.dirSep..prevPath

                                                    --the standard path naming has a priority
                                                    if n < 8 or tryPath:sub(n - 7,n - 1) ~= ".layers" then
                                                        path = basePath..prevPath..item
                                                        tryFile = io.open(path,"r")
                                                        if tryFile then
                                                            tryFile:close()
                                                            break
                                                        end
                                                    end
                                                    n = n - 1
                                                end
                                                if n <= 0 then
                                                    path = oldPath..item
                                                end
                                            end
                                        else
                                            path = oldPath..item
                                        end
                                        imageLayer:SetSourceImage(path)
                                        imageLayer:SetName(item)
                                    end
                                    layers[pos].switch:SetValue(frame, item)
                                end
                            end
                            pos = pos + 1
                        end
                        frame = frame + 1
                    end
                end
            end
        end
    end
    f:close()

    if rootGroup ~= nil then
        moho:SetSelLayer(rootGroup)
        if setDoc == 0 then
            moho.document:SetStartFrame(1)
            moho.document:SetEndFrame(frame)
        end
    end
end
